from ._internal.frameworks.easyocr import get
from ._internal.frameworks.easyocr import get_runnable
from ._internal.frameworks.easyocr import load_model
from ._internal.frameworks.easyocr import save_model

__all__ = ["load_model", "save_model", "get", "get_runnable"]
